package corefdb

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/lazywriter"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl/filepack"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/hashmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"

	"github.com/tidwall/sjson"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/utiliymdl/guidmdl"

	"github.com/tidwall/buntdb"
	"github.com/tidwall/gjson"
)

const (
	fileStatusReady                     = 0
	fileStatusUpdatingData              = 1
	fileStatusUpdatingIndex             = 2
	fileStatusOffsetInFile              = 0
	isReadyForUploadOffsetInFile        = 1
	isUpdatedAndNotCommitedOffsetInFile = 2
	isReorgRequiredOffsetInFile         = 3
	isReindexRequiredOffsetInFile       = 4
	footerOffsetInFile                  = 5
	footerOffsetReservedSize            = 15
	footerSizeOffset                    = 20
	filehashOffest                      = 35
	lastUpdatedOffset                   = 43
	dataStartOffset                     = 53
	sizeReservedForHeaders              = 53
	lineBreak                           = "\r\n"

	// BucketTypeSimple -
	BucketTypeSimple = "Normal"
	// BucketTypePack -
	BucketTypePack = "Pack"
	// BucketTypeAppend -
	BucketTypeAppend = `Append`
	// IndexKeyValSeperator -
	IndexKeyValSeperator = "="
	// FileType - represents key for type of file. Used whenever we need to set file type field in json
	FileType         = "fileType"
	FileTypeFDBIndex = "FDBIndex"
	// GetAll - wraper gjson query to return all results.
	GetAll = "#[*]"
)

// ErrNoDataFound - This error describes that the required data might be deleted and not found. Kidly ignore this error in caller.
var ErrNoDataFound = errors.New("No data found")

// Bucket - Bucket
type Bucket struct {
	BucketID        string                 `json:"bucketId"`
	IsDynamicName   bool                   `json:"isDynamicName"`
	BucketNameQuery string                 `json:"bucketNameQuery"`
	Indices         []string               `json:"indices"`
	BucketPath      string                 `json:"bucketPath"`
	InFileIndexMap  map[string]InFileIndex `json:"inFileIndexMap"`
	BucketType      string                 `json:"bucketType"`
}

// InFileIndex -
type InFileIndex struct {
	FileType    string       `json:"fileType"`
	IndexFields []IndexField `json:"indexFields"`
}

// Index - Index
type Index struct {
	indexObj       *buntdb.DB
	IndexID        string       `json:"indexId"`
	IndexNameQuery string       `json:"indexNameQuery"`
	BucketSequence []string     `json:"bucketSequence"`
	IndexFields    []IndexField `json:"indexFields"`
	IsDynamicName  bool         `json:"isDynamicName"`
}

// IndexField - IndexField
type IndexField struct {
	FieldName string `json:"fieldName"`
	Query     string `json:"query"`
}

// GetNewIndex - GetNewIndex returns new index
func (fdb *FDB) GetNewIndex(indexNameQuery string, IsDynamicName bool) (*Index, error) {
	fdb.indexMux.Lock()
	defer fdb.indexMux.Unlock()
	if _, ok := fdb.indices[indexNameQuery]; ok {
		return nil, errormdl.Wrap("Index name already found")
	}
	idx := &Index{
		IndexNameQuery: indexNameQuery,
		IndexID:        indexNameQuery,
		IsDynamicName:  IsDynamicName,
	}
	db, err := fdb.GetIndexDB(idx)
	if err != nil {
		return nil, err
	}
	idx.indexObj = db

	fdb.indices[idx.IndexID] = idx
	return idx, nil
}

// RegisterNewIndex - RegisterNewIndex returns new index
func (fdb *FDB) RegisterNewIndex(idx *Index) (*Index, error) {
	fdb.indexMux.Lock()
	defer fdb.indexMux.Unlock()
	if _, ok := fdb.indices[idx.IndexID]; ok {
		return nil, errormdl.Wrap("Index ID already found")
	}

	db, err := fdb.GetIndexDB(idx)
	if err != nil {
		return nil, err
	}
	idx.indexObj = db

	fdb.indices[idx.IndexID] = idx
	return idx, nil
}

// GetIndexDB -  GetIndexDB
func (fdb *FDB) GetIndexDB(index *Index) (*buntdb.DB, error) {
	// dbPath := filepath.Join(fdb.DBPath, INDEXFOLDER)
	// filemdl.CreateDirectoryRecursive(dbPath)
	// dbPath = filepath.Join(dbPath, index.IndexID)
	db, err := buntdb.Open(":memory:")
	if err != nil {
		return nil, err
	}
	return db, nil
}

// GetNewBucket - return Bucket Obj
func (fdb *FDB) GetNewBucket(BucketNameQuery string, IsDynamicName bool, parentBucket *Bucket) *Bucket {
	fdb.bLocker.Lock()
	defer fdb.bLocker.Unlock()
	bucket := &Bucket{
		BucketID:        guidmdl.GetGUID(),
		BucketNameQuery: BucketNameQuery,
		IsDynamicName:   IsDynamicName,
	}
	if IsDynamicName {
		BucketNameQuery = "$$" + BucketNameQuery
	}
	bucket.BucketPath = filepath.Join(parentBucket.BucketPath, BucketNameQuery)
	fdb.buckets[bucket.BucketID] = bucket
	return bucket
}

// NewBucket - return Bucket Obj
func (fdb *FDB) NewBucket(bucket, parentBucket *Bucket) *Bucket {
	fdb.bLocker.Lock()
	defer fdb.bLocker.Unlock()
	// bucket := &Bucket{
	// 	BucketID:        guidmdl.GetGUID(),
	// 	BucketNameQuery: BucketNameQuery,
	// 	IsDynamicName:   IsDynamicName,
	// }
	bucketNameQuery := bucket.BucketNameQuery
	if bucket.IsDynamicName {
		bucketNameQuery = "$$" + bucket.BucketNameQuery
	}
	bucket.BucketPath = filepath.Join(parentBucket.BucketPath, bucketNameQuery)
	fdb.buckets[bucket.BucketID] = bucket
	return bucket
}

// // SetIsBucketPackType - SetIsBucketPackType
// func (bucket *Bucket) SetIsBucketPackType(isPackType bool) {
// 	bucket.IsSinglePackType = isPackType
// }

// SetBucketType -
func (bucket *Bucket) SetBucketType(bucketType string) {
	bucket.BucketType = bucketType
}

// SetInFileIndex - SetInFileIndex
func (bucket *Bucket) SetInFileIndex(inFileIndex InFileIndex) {
	if bucket.InFileIndexMap == nil {
		bucket.InFileIndexMap = make(map[string]InFileIndex)
	}
	bucket.InFileIndexMap[inFileIndex.FileType] = inFileIndex
}

// SetInFileIndexes - SetInFileIndexes
func (bucket *Bucket) SetInFileIndexes(inFileIndexes []InFileIndex) {
	if bucket.InFileIndexMap == nil {
		bucket.InFileIndexMap = make(map[string]InFileIndex, len(inFileIndexes))
	}
	for _, inFileIndex := range inFileIndexes {
		bucket.InFileIndexMap[inFileIndex.FileType] = inFileIndex
	}
}

// SetBucket - set bucket in index
func (index *Index) SetBucket(bucket *Bucket) *Index {
	if bucket != nil {
		index.BucketSequence = append(index.BucketSequence, bucket.BucketID)
		bucket.Indices = append(bucket.Indices, index.IndexID)
	}
	return index
}

// SetFields - SetFields
func (index *Index) SetFields(indexField ...IndexField) *Index {
	index.IndexFields = indexField
	return index
}

// bucketNameResolver - returns bucket name
func (fdb *FDB) bucketNameResolver(bucketID string, rs *gjson.Result) (string, error) {
	fdb.bLocker.Lock()
	defer fdb.bLocker.Unlock()
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return "", errormdl.Wrap("Bucket not found: " + bucketID)
	}
	name := bucket.BucketNameQuery
	if bucket.IsDynamicName {
		name = rs.Get(name).String()
	}
	if name == "" {
		return name, errormdl.Wrap("Bucket name should not be empty: " + bucket.BucketNameQuery)
	}
	return name, nil
}

// resolveIndex- returns bucket path by reolving index
func (fdb *FDB) resolveIndex(index *Index, rs *gjson.Result) (string, error) {
	path := ""
	for _, bucketID := range index.BucketSequence {
		bucketName, err := fdb.bucketNameResolver(bucketID, rs)
		if err != nil {
			return path, err
		}
		path = filepath.Join(path, bucketName)
	}
	indexName := index.IndexNameQuery
	if index.IsDynamicName {
		indexName = rs.Get(index.IndexNameQuery).String()
	}
	if indexName == "" {
		return path, errormdl.Wrap("required attribute not provided:" + index.IndexNameQuery)
	}
	path = filepath.Join(path, indexName)
	return path, nil
}

// createIndexJSON - create JSON with index field data
func createIndexJSON(index *Index, rs *gjson.Result) (string, error) {
	json := `{}`
	for _, indexField := range index.IndexFields {
		if rs.Get(indexField.Query).Value() == nil {
			return json, errormdl.Wrap("please provide value for index field: " + indexField.Query)
		}
		json, _ = sjson.Set(json, indexField.FieldName, rs.Get(indexField.Query).Value())
	}
	return json, nil
}

// updateIndexJSON - update JSON with index field data
func updateIndexJSON(index *Index, existingData string, rs *gjson.Result) (string, error) {
	json := existingData
	var err error
	for _, indexField := range index.IndexFields {
		if rs.Get(indexField.Query).Value() == nil {
			continue
		}
		json, err = sjson.Set(json, indexField.FieldName, rs.Get(indexField.Query).Value())
		loggermdl.LogError("Error - ", err)
	}
	return json, nil
}

// createIndexJSON - create JSON with index field data
func createIndexJSONByFields(indexFields []IndexField, rs *gjson.Result) (string, error) {
	json := `{}`
	for _, indexField := range indexFields {
		val := rs.Get(indexField.Query).Value()
		// validation
		if val == nil {
			return "", errormdl.Wrap("please provide value for index field: " + indexField.Query)
		}
		json, _ = sjson.Set(json, indexField.FieldName, val)
	}
	return json, nil
}

// UpdateIndexLazyObjectInCache - updates index data in lay writer cache
func UpdateIndexLazyObjectInCache(indexID string, indexData *Index) error {
	// lazy cache must be present for provided indexID
	lazyObj, ok := IndexLazyObjHolder.Get(indexID)
	if !ok {
		loggermdl.LogError("index not found in lazy writer cache")
		return errormdl.Wrap("index not found in lazy writer cache")
	}

	idxLazyData, ok := lazyObj.(lazywriter.LazyCacheObject)
	if !ok {
		loggermdl.LogError("interface type is not lazywriter.LazyCacheObject")
		return errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
	}

	// idxLazyData.GJSONData = indexData
	idxLazyData.InterfaceData = indexData
	if ok := IndexMaster.SaveOrUpdateDataInCache(idxLazyData); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return errormdl.Wrap("failed to update index data in lazy cache")
	}

	IndexLazyObjHolder.SetNoExpiration(indexID, idxLazyData)
	return nil
}

func getUpdatedFilePath(filePath string, data gjson.Result) (string, error) {
	updatedFilePath := ""
	filePathSplit := strings.Split(filePath, string(filepath.Separator))
	for _, pathChunk := range filePathSplit {
		if strings.Index(pathChunk, "$$") != -1 {
			indexFieldQuery := strings.TrimPrefix(pathChunk, "$$")
			value := data.Get(indexFieldQuery)
			if value.Exists() {
				updatedFilePath = filepath.Join(updatedFilePath, value.String())
			} else {
				return "", errormdl.Wrap("index filed value not provided: " + indexFieldQuery)
			}
		} else {
			updatedFilePath = filepath.Join(updatedFilePath, pathChunk)
		}

	}

	return updatedFilePath, nil
}

// AppendDataInLazyObjectInCache - updates index data in lay writer cache
func AppendDataInLazyObjectInCache(bucketID string, data gjson.Result) error {
	// lazy cache must be present for provided bucketID
	// loggermdl.LogError("AppendDataInLazyObjectInCache")
	mutex := sync.Mutex{}
	mutex.Lock()
	lazyObj, ok := AppendLazyObjHolder.Get(bucketID)
	if !ok {
		loggermdl.LogError(AppendLazyObjHolder.GetItems())
		return errormdl.Wrap("lazyObject not found in lazywriter cache")
	}

	bucketLazyData, ok := lazyObj.(lazywriter.LazyCacheObject)
	if !ok {
		loggermdl.LogError("interface type is not lazywriter.LazyCacheObject")
		return errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
	}

	filePath := bucketLazyData.FileName
	if filePath == "" {
		loggermdl.LogError("filepath not found in lazywriter")
		return errormdl.Wrap("filepath not found in lazywriter")

	}
	updatedFilePath, err := getUpdatedFilePath(filePath, data)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	bucketLazyData.FileName = updatedFilePath
	updatedData := ""
	if bucketLazyData.InterfaceData == nil {
		updatedData = data.String()
	} else {
		previousData, ok := bucketLazyData.InterfaceData.(string)
		if !ok {
			return errormdl.Wrap("previous lazy data is not of type string")
		}
		updatedData = previousData + string(lineBreak) + data.String()
	}

	// bucketLazyData.GJSONData = indexData
	bucketLazyData.InterfaceData = updatedData
	if ok := AppendMaster.SaveOrUpdateDataInCache(bucketLazyData); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return errormdl.Wrap("failed to update index data in lazy cache")
	}

	AppendLazyObjHolder.SetNoExpiration(bucketID, bucketLazyData)
	mutex.Unlock()
	return nil
}

// searchDataInPack - searchDataInPack
func searchDataInPack(paths []string, searchTillPath string, restoreFileFromPack bool) (*gjson.Result, error) {
	dirPath, _ := filepath.Split(paths[0])
	packPath, err := filepack.SearchPackFilePath(dirPath, searchTillPath)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	if len(packPath) == 0 {
		return &gjson.Result{}, nil
	}
	fileDataMap, err := filepack.GetFileDataListFromPack(packPath, paths)

	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	dataString := "[]"
	for fileName, dataByte := range fileDataMap {
		if len(dataByte) > 0 {
			dataString, _ = sjson.Set(dataString, "-1", gjson.Parse(string(dataByte)).Value())
			if restoreFileFromPack {
				restoreFilesFromFilePack(fileName, dataByte)
			}
		}
	}
	data := gjson.Parse(dataString)

	return &data, nil
}

// restoreFilesFromFilePack - restoreFilesFromFilePack
func restoreFilesFromFilePack(fileName string, fileData []byte) error {
	if len(fileData) > 0 {
		err := filemdl.WriteFile(fileName, fileData, true, false)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return errormdl.CheckErr(err)
		}
	}
	return nil
}

// readDataFile - reads file and returns file data
func (fdb *FDB) readDataFile(path []string) (*gjson.Result, error) {
	sb := strings.Builder{}
	sb.WriteString("[")
	for i := 0; i < len(path); i++ {
		updatedPath := filepath.Join(fdb.DBPath, path[i])
		ba := []byte{'{', '}'}
		var err error
		if filemdl.FileAvailabilityCheck(updatedPath) {
			ba, err = filemdl.ReadFile(updatedPath)
			if err != nil {
				loggermdl.LogError(err)
				return nil, err
			}

			// loggermdl.LogError("File data for index = ", i, " - ", string(ba))
			_, err = sb.Write(ba)
			if err != nil {
				loggermdl.LogError(err)
				return nil, err
			}
			sb.WriteString(",")
		}
	}
	sb.WriteString("]")
	finalResult := strings.Replace(sb.String(), ",]", "]", 1)
	rs := gjson.Parse(finalResult)
	return &rs, nil
}

// GetDataByConcat - GetDataByConcat
func GetDataByConcat(dbName, indexID string, query ...string) (*gjson.Result, error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return nil, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return nil, errormdl.Wrap("INDEX not found: " + indexID)
	}
	// path := []string{}
	sb := strings.Builder{}
	sb.WriteString("[")
	index.indexObj.View(func(tx *buntdb.Tx) error {

		tx.Ascend(indexID, func(key, value string) bool {
			sb.WriteString((value + ","))
			return true
		})

		return nil
	})
	sb.WriteString("]")
	finalData := strings.Replace(sb.String(), ",]", "]", 1)
	rsJSON := gjson.Parse(finalData)
	for i := 0; i < len(query); i++ {
		rsJSON = rsJSON.Get(query[i] + "#")
		if rsJSON.Value() == nil {
			break
		}
	}

	return nil, nil
}

func saveDataInNormalBucket(filePath string, rs *gjson.Result) error {
	// fileLock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(fileLock)
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return errormdl.CheckErr(err)
	// }
	err := filemdl.WriteFile(filePath, []byte(rs.String()), true, false)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	return nil
}

func saveDataInPackBucket(fdb *FDB, filePath string, index *Index, rs *gjson.Result) error {
	// check is file locked
	// if yes
	// wait
	// else
	// /*  */update filestatus to fileStatusUpdatingData
	// compress and encrypt data
	// calculate filehash
	//save data in file & getoffset and len
	//update footer address
	// update filestatuse to fileStatusUpdatingIndex
	// update index
	// update filestatus to fileStatusReady

	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket := fdb.buckets[bucketID]
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return errormdl.Wrap("please specify fileType")
	}
	isFilePresent := filemdl.FileAvailabilityCheck(filePath)
	if !isFilePresent {
		dir, _ := filepath.Split(filePath)
		err := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return errormdl.CheckErr(err)
		}

		err = initializeWithHeader(filePath)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}

		err = addFDBIndexEntryFile(filePath, bucket, index.IndexFields, rs)
		if err != nil {
			loggermdl.LogError("fail to add fdb index entry in file: ", err)
			return errormdl.Wrap("fail to write data in file")
		}
	}
	return addDataToSpecialBucketFile(filePath, bucket, rs)
}

func saveDataInAppendBucket(filePath string, rs *gjson.Result, bucket *Bucket) error {

	// loggermdl.LogError("isLazyWriterEnabled", isLazyWriterEnabled)
	if isLazyWriterEnabled {
		err := AppendDataInLazyObjectInCache(bucket.BucketID, *rs)
		return err
	}
	//  else
	_, _, err := filemdl.AppendDataToFile(filePath, []byte(rs.String()+lineBreak), true)
	return err

}

func addFDBIndexEntryFile(filePath string, bucket *Bucket, infileIndexFields []IndexField, rs *gjson.Result) error {
	// adding FDBIndex as infileIndex
	FDBIndexInFileIndexMap := InFileIndex{
		FileType:    "FDBIndex",
		IndexFields: infileIndexFields,
	}
	bucket.SetInFileIndex(FDBIndexInFileIndexMap)
	// create infile index data
	newData, err := createIndexJSONByFields(infileIndexFields, rs)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	newData, _ = sjson.Set(newData, "fileType", "FDBIndex")
	newDataObj := gjson.Parse(newData)
	return addDataToSpecialBucketFile(filePath, bucket, &newDataObj)
}

func addDataToSpecialBucketFile(filePath string, bucket *Bucket, rs *gjson.Result) error {
	fileType := rs.Get("fileType").String()
	inFileIndexForFileType, ok := bucket.InFileIndexMap[fileType]
	if !ok {
		return errormdl.Wrap("infileIndex for specified fileType not found")
	}
	// fileLock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(fileLock)

	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return errormdl.CheckErr(err)
	// }

	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer f.Close()
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	// if file is being created for first time
	fileStatus := fileStatusReady
	// if isFilePresent {
	fileStatus, err = getFileStatus(f)
	if err != nil {
		if err.Error() == "EOF" {
			fileStatus = fileStatusReady
		} else {
			loggermdl.LogError(err)
			return err
		}
	}
	// }
	if fileStatusReady == fileStatus {

		// if its first write in file then add fdb index data of fileType FDBIndex
		//prepare data
		indexRowJSON, err := createIndexJSONByFields(inFileIndexForFileType.IndexFields, rs)
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", inFileIndexForFileType.FileType)
		fileHash, err := securitymdl.GetHash(indexRowJSON)
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
		// compress & encrypt data
		previousIndexData, err := getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		// loggermdl.LogDebug("previous index data", previousIndexData)
		footerStartOffset := getFooterOffset(f)
		if footerStartOffset == -1 {
			loggermdl.LogError("fail to fetch infile index offset")
			return errormdl.Wrap("fail to fetch infile index data")
		}
		dataString := rs.String()
		err = setFileStatusFlag(f, fileStatusUpdatingData)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		// write data
		dataSize, err := addFileDataInFile(f, footerStartOffset, dataString, true)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
		// append new entry in infile index
		parsedindexRowJSON := gjson.Parse(indexRowJSON)
		// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
		updatedIndexData, _ := sjson.Set(previousIndexData, "-1", parsedindexRowJSON.Value())
		// loggermdl.LogDebug("updatedIndexData", updatedIndexData)
		// updating infile index
		err = setFileStatusFlag(f, fileStatusUpdatingIndex)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}

		footerNewOffset := footerStartOffset + dataSize
		err = setFooterOffset(f, footerNewOffset)
		err = setFooterSize(f, int64(len(updatedIndexData)))
		err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
		err = setFileStatusFlag(f, fileStatusReady)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		f.Sync()
	} else {
		// retry after timeout
	}

	return nil

}
func getInFileIndexData(f *os.File) (string, error) {
	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return "[]", errormdl.Wrap("fail to fetch infile index data")
	}
	footerSize, err := getFooterSize(f)
	if err != nil {
		return "[]", nil
	}
	if footerSize == 0 {
		return "[]", nil
	}
	dataByte, err := filemdl.ReadFileFromOffset(f, footerStartOffset, footerSize)
	if err != nil {
		if err.Error() == "EOF" {
			loggermdl.LogError("EOF")
			return "[]", nil
		}
		loggermdl.LogError("error while fetching index data", err)
		return "[]", err
	}
	return string(dataByte), nil
}

func updateDataInNormalBucket(fdb *FDB, filePaths []string, rs *gjson.Result) (*gjson.Result, []string, []error) {
	resultStr := "[]"
	errList := []error{}
	filesUpdated, updatedCnt := make([]string, len(filePaths)), 0
	for _, filePath := range filePaths {
		path := filepath.Join(fdb.DBPath, filePath)
		data, err := updateFileData(path, *rs)
		if err != nil {
			// loggermdl.LogError("failed to update data in file: " + path + " : " + err.Error())
			errList = append(errList, errormdl.Wrap("failed to update data in file: "+path+" : "+err.Error()))
			continue
		}
		if filePath != "" {
			// filesUpdated = append(filesUpdated, filePath)
			filesUpdated[updatedCnt] = filePath
			updatedCnt++
			resultStr, _ = sjson.Set(resultStr, "-1", data.Value())
		} else {
			loggermdl.LogError("invalid path for update - ", filePath)
		}
	}
	result := gjson.Parse(resultStr)
	return &result, filesUpdated[:updatedCnt], errList
}

func updateFileData(filePath string, rs gjson.Result) (*gjson.Result, error) {
	// fileLock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(fileLock)
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return nil, errormdl.CheckErr(err)
	// }
	data, err := filemdl.FastReadFile(filePath)
	if err != nil {
		return nil, err
	}
	existingDataStr := string(data)
	rs.ForEach(func(key, val gjson.Result) bool {
		// updating existing data
		existingDataStr, _ = sjson.Set(existingDataStr, key.String(), val.Value())
		return true
	})

	err = filemdl.FastWriteFile(filePath, []byte(existingDataStr), true, false, true)
	if err != nil {
		return nil, err
	}
	updatedData := gjson.Parse(existingDataStr)
	return &updatedData, nil
}

func updateDataInFileIndexBucket(fdb *FDB, bucket *Bucket, filePaths []string, rs *gjson.Result, infileIndexQuery []string) (*gjson.Result, []string, []error) {
	finalResultArray := []gjson.Result{}
	errList := []error{}
	dataUpdatedAtFilePaths := make([]string, len(filePaths))
	dataUpdatedAtFilePathsCnt := 0
	for _, path := range filePaths {
		filePath := filepath.Join(fdb.DBPath, path)

		resultArray, err := updateDataWithInFileIndex(filePath, bucket, rs, infileIndexQuery)
		if err != nil {
			errList = append(errList, err)
			continue
		}

		// dataUpdatedAtFilePaths = append(dataUpdatedAtFilePaths, path)
		dataUpdatedAtFilePaths[dataUpdatedAtFilePathsCnt] = path
		dataUpdatedAtFilePathsCnt++
		finalResultArray = append(finalResultArray, resultArray.Array()...)
	}

	resultListStr := "[]"
	for _, resultObj := range finalResultArray {
		resultListStr, _ = sjson.Set(resultListStr, "-1", resultObj.Value())
	}
	result := gjson.Parse(resultListStr)
	return &result, dataUpdatedAtFilePaths[:dataUpdatedAtFilePathsCnt], errList
}

// updateDataWithInFileIndex - updateDataWithInFileIndex
func updateDataWithInFileIndex(filePath string, bucket *Bucket, rs *gjson.Result, infileIndexQueries []string) (*gjson.Result, error) {
	// check fileType index availability
	// check is data present
	// if data present
	// then calculate size of updated data
	// if size is less than or equal to previuos data size
	// then write at the same location
	// else if size of updated data is more than existing data then append it to end of data
	// update startOffset and data size of file in index row
	// update footer offset and footer size
	requestedFileType := rs.Get("fileType").String()
	_, ok := bucket.InFileIndexMap[requestedFileType]
	if !ok {
		return nil, errormdl.Wrap("infileIndex for specified fileType not found")
	}
	if !filemdl.FileAvailabilityCheck(filePath) {
		return nil, errormdl.Wrap("file not found: " + filePath)
	}
	// lock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(lock)
	// if err != nil {
	// 	return nil, errormdl.Wrap("fail to update data" + err.Error())
	// }
	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer f.Close()
	if err != nil {
		return nil, errormdl.Wrap("fail to update data" + err.Error())
	}
	indexDataString, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	infileIndexQueries = append(infileIndexQueries, `#[fileType==`+requestedFileType+`]`)
	indexRows := gjson.Parse(indexDataString)
	indexRecordsToUpdate := indexRows
	for _, query := range infileIndexQueries {
		if query == GetAll {
			// return all results
			break
		}
		indexRecordsToUpdate = indexRecordsToUpdate.Get(query + "#")
	}
	indexRecordsToUpdateObjs := indexRecordsToUpdate.Array()
	if len(indexRecordsToUpdateObjs) == 0 {
		return nil, ErrNoDataFound
	}
	// updating first record
	resultArrayStr := "[]"
	for _, recordToUpdateIndexRow := range indexRecordsToUpdateObjs {
		// loggermdl.LogDebug("recordToUpdateIndexRow", recordToUpdateIndexRow)
		result, err := updateDataInFileIndex(f, bucket, recordToUpdateIndexRow, *rs, indexRows)
		if err != nil {
			return nil, errormdl.Wrap("fail to update data" + err.Error())
		}
		resultArrayStr, _ = sjson.Set(resultArrayStr, "-1", result.Value())
	}
	resultData := gjson.Parse(resultArrayStr)
	return &resultData, nil
}

func updateDataInFileIndex(f *os.File, bucket *Bucket, recordToUpdateIndexRow gjson.Result, rs, existingIndexRows gjson.Result) (*gjson.Result, error) {
	fileStartOffset := recordToUpdateIndexRow.Get("startOffset").Int()
	dataSize := recordToUpdateIndexRow.Get("dataSize").Int()
	if fileStartOffset == 0 || dataSize == 0 {
		loggermdl.LogError("index row details incorrect - start offset :", fileStartOffset, " data size :", dataSize)
		return nil, errormdl.Wrap("index row details incorrect")
	}
	// compare data size
	// TODO: compare with encrypted data

	// update fields data
	// existingData, err := filemdl.ReadFileFromOffset(f, fileStartOffset, dataSize)
	existingData, err := getFileDataFromFile(f, fileStartOffset, dataSize)
	if err != nil {
		return nil, err
	}

	updatedDataStr := strings.TrimSpace(string(existingData))
	rs.ForEach(func(key, val gjson.Result) bool {
		// updating existing data
		updatedDataStr, _ = sjson.Set(updatedDataStr, key.String(), val.Value())
		return true
	})
	newDataSize := int64(len(updatedDataStr))
	footerStartOffset := getFooterOffset(f)
	updatedFooterOffset := footerStartOffset
	err = setFileStatusFlag(f, fileStatusUpdatingData)
	if err != nil {
		return nil, err
	}

	if len(strings.TrimSpace(updatedDataStr)) <= len(strings.TrimSpace(string(existingData))) {
		newDataSize, err = addFileDataInFile(f, fileStartOffset, updatedDataStr, false)
		if err != nil {
			return nil, err
		}
		// newDataSize = newDataSize + 2
	} else {

		newDataSize, err = addFileDataInFile(f, footerStartOffset, updatedDataStr, true)
		updatedFooterOffset = footerStartOffset + newDataSize
		fileStartOffset = footerStartOffset
	}
	// indexRows, err := getInFileIndexData(f)
	// if err != nil {
	// 	return nil, err
	// }
	// update startofffset and datasize in infile index
	// updateIndexRow
	updatedIndexRowStr := recordToUpdateIndexRow.String()

	recordToUpdateIndexRow.ForEach(func(key, value gjson.Result) bool {
		indexFieldKey := key.String()
		if rs.Get(indexFieldKey).Exists() {
			updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, indexFieldKey, rs.Get(indexFieldKey).Value())
		}
		return true
	})
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "startOffset", fileStartOffset)
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "dataSize", newDataSize)

	updatedIndexRows, err := updateIndexRow(existingIndexRows, recordToUpdateIndexRow, gjson.Parse(updatedIndexRowStr))

	if err != nil {
		return nil, err
	}

	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return nil, err
	}
	// loggermdl.LogDebug("updated index data", updatedIndexRows)
	err = setIndexDataInFile(f, updatedFooterOffset, updatedIndexRows.String())
	if err != nil {
		return nil, err
	}
	err = setFooterOffset(f, updatedFooterOffset)
	if err != nil {
		return nil, err
	}
	err = setFooterSize(f, int64(len(updatedIndexRows.String())))
	if err != nil {
		return nil, err
	}

	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return nil, err
	}
	f.Sync()
	// TODO: discussion
	// err = setFileReadyForUploadFlag(f, true)
	// if err != nil {
	// 	return err
	// }
	updatedData := gjson.Parse(updatedDataStr)
	return &updatedData, nil
}

func updateIndexRow(indexRows gjson.Result, previousIndexRow gjson.Result, updatedRow gjson.Result) (gjson.Result, error) {
	indexRowObjs := indexRows.Array()
	if len(indexRowObjs) == 0 {
		return indexRows, errormdl.Wrap("no data found to update")
	}
	// loggermdl.LogDebug("indexRows", indexRows)

	prevIndexRowString := previousIndexRow.String()
	foundRowIndexes := []int{}

	for index, indexRowObj := range indexRowObjs {
		if indexRowObj.String() != "" && prevIndexRowString != "" && indexRowObj.String() == prevIndexRowString {
			foundRowIndexes = append(foundRowIndexes, index)
			break
		}
	}

	if len(foundRowIndexes) == 0 {
		return indexRows, errormdl.Wrap("no record found to update")
	}
	var err error
	updatedIndexDataString := indexRows.String()
	for _, foundRowIndex := range foundRowIndexes {
		updatedIndexDataString, err = sjson.Set(updatedIndexDataString, strconv.Itoa(foundRowIndex), updatedRow.Value())
		if err != nil {
			loggermdl.LogError(err)
		}
	}
	return gjson.Parse(updatedIndexDataString), nil
}

func initializeWithHeader(filePath string) error {
	// fileLock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(fileLock)
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return errormdl.CheckErr(err)
	// }
	// initailize file with headers and fdb index entry
	f, err := filemdl.OpenFile(filePath, os.O_CREATE|os.O_RDWR|os.O_SYNC, 0777)
	defer f.Close()
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	_, err = f.WriteAt([]byte(strconv.Itoa(fileStatusReady)), fileStatusOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// isFile ready for upload  =0
	_, err = f.WriteAt([]byte("0"), isReadyForUploadOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isUpdatedAndNotCommitedOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReorgRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReindexRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// _, err = f.WriteAt([]byte(appendPaddingToNumber(sizeReservedForHeaders, 15)), footerOffsetInFile)
	err = setFooterOffset(f, sizeReservedForHeaders+int64(len(lineBreak)))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	err = setFooterSize(f, 0)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("filehash"), filehashOffest)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	timestamp := strconv.FormatInt(time.Now().Unix(), 10)
	_, err = f.WriteAt([]byte(timestamp), lastUpdatedOffset)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("\r\n"), sizeReservedForHeaders)
	return err
}

func appendPaddingToNumber(value int64, padNumber int) string {
	no := strconv.Itoa(padNumber)
	return fmt.Sprintf("%0"+no+"d", value)
}

func getFileStatus(f *os.File) (int, error) {

	data, err := filemdl.ReadFileFromOffset(f, fileStatusOffsetInFile, 1)
	if err != nil {
		loggermdl.LogError(err)
		return -1, err
	}
	status, err := strconv.Atoi(string(data))
	return status, err
}

func getFooterOffset(f *os.File) int64 {
	data, err := filemdl.ReadFileFromOffset(f, footerOffsetInFile, footerOffsetReservedSize)
	if err != nil {
		loggermdl.LogError(err)
		return -1
	}
	footerOffset, err := strconv.Atoi(string(data))
	if err != nil {
		loggermdl.LogError("err", err)
		return -1
	}

	return int64(footerOffset)
}

func setFileStatusFlag(f *os.File, fileStatus int) error {
	status := strconv.Itoa(fileStatus)
	_, err := filemdl.WriteFileAtOffset(f, fileStatusOffsetInFile, []byte(status))
	return err
}

func setFileReadyForUploadFlag(f *os.File, isReadyToUpload bool) error {
	flagVal := strconv.FormatBool(isReadyToUpload)
	_, err := filemdl.WriteFileAtOffset(f, isReadyForUploadOffsetInFile, []byte(flagVal))
	return err
}

func setFileUpdatedAndNotCommitedFlag(f *os.File, isUpdatedAndNotCommited bool) error {
	flagVal := strconv.FormatBool(isUpdatedAndNotCommited)
	_, err := filemdl.WriteFileAtOffset(f, isUpdatedAndNotCommitedOffsetInFile, []byte(flagVal))
	return err
}

func setFileReorgRequiredFlag(f *os.File, isReorgRequired bool) error {
	flagVal := strconv.FormatBool(isReorgRequired)
	_, err := filemdl.WriteFileAtOffset(f, isReorgRequiredOffsetInFile, []byte(flagVal))
	return err
}

func setFileReindexRequiredFlag(f *os.File, isReindexRequired bool) error {
	flagVal := strconv.FormatBool(isReindexRequired)
	_, err := filemdl.WriteFileAtOffset(f, isReindexRequiredOffsetInFile, []byte(flagVal))
	return err
}

func setFooterOffset(f *os.File, footerOffset int64) error {
	footerOffestInString := appendPaddingToNumber(footerOffset, 15)
	_, err := filemdl.WriteFileAtOffset(f, footerOffsetInFile, []byte(footerOffestInString))
	return err
}

func setFooterSize(f *os.File, footerSize int64) error {
	footerSizeInString := appendPaddingToNumber(footerSize, 15)
	_, err := filemdl.WriteFileAtOffset(f, footerSizeOffset, []byte(footerSizeInString))
	return err
}

func getFooterSize(f *os.File) (int64, error) {
	data, err := filemdl.ReadFileFromOffset(f, footerSizeOffset, 15)
	if err != nil {
		return -1, err
	}
	footerSize, err := strconv.Atoi(string(data))
	if err != nil {
		loggermdl.LogError("err", err)
		return -1, err
	}

	return int64(footerSize), nil
}

func setIndexDataInFile(f *os.File, footerOffset int64, indexData string) error {
	_, err := filemdl.WriteFileAtOffset(f, footerOffset, []byte(indexData))
	return err
}

func compressData(data []byte) ([]byte, error) {
	var hashError error
	dataByteToWriteRes, hashError := filemdl.ZipBytes(data)
	if errormdl.CheckErr2(hashError) != nil {
		return data, errormdl.CheckErr2(hashError)
	}
	return dataByteToWriteRes, nil
}

func decompressData(data []byte) ([]byte, error) {
	var hashError error
	dataByteToWriteRes, hashError := filemdl.UnZipBytes(data)
	if errormdl.CheckErr2(hashError) != nil {
		return data, errormdl.CheckErr2(hashError)
	}
	return dataByteToWriteRes, nil
}

func encryptData(data, key []byte) (dataOut []byte, err error) {
	dataOut, err = compressData(data)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	dataOut, err = securitymdl.AESEncrypt(dataOut, key)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	return
}

func decryptData(data, key []byte) (dataOut []byte, err error) {

	dataOut, err = securitymdl.AESDecrypt(data, key)
	if errormdl.CheckErr1(err) != nil {
		return
	}

	dataOut, err = decompressData(dataOut)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	return
}

func addFileDataInFile(f *os.File, offset int64, data string, breakLine bool) (int64, error) {
	var err error
	dataBytes := []byte(data)

	if isSecurityEnabled {
		keyBytes, err := GetKeyWithFileNameAndDefaultKey(f.Name())
		if errormdl.CheckErr(err) != nil {
			return 0, errormdl.CheckErr(err)
		}
		dataBytes, err = encryptData(dataBytes, keyBytes)
		if errormdl.CheckErr(err) != nil {
			return 0, errormdl.CheckErr(err)
		}
	}
	if !isSecurityEnabled && isCompressionEnabled {
		dataBytes, err = filemdl.ZipBytes(dataBytes)
		if err != nil {
			return 0, err
		}
	}
	if breakLine {
		dataBytes = append(dataBytes, []byte(lineBreak)...)
	}
	dataSize, err := filemdl.WriteFileAtOffset(f, offset, dataBytes)
	return dataSize, err
}

func getFileDataFromFile(f *os.File, startOffset, dataSize int64) ([]byte, error) {

	dataByte, err := filemdl.ReadFileFromOffset(f, startOffset, dataSize)

	if isSecurityEnabled {
		keyBytes, err := GetKeyWithFileNameAndDefaultKey(f.Name())
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError("GetKeyWithFileNameAndDefaultKey", err)
			return dataByte, err
		}
		dataByte, err = decryptData(dataByte, keyBytes)
		if err != nil {
			loggermdl.LogError("decryptData", err)
			return dataByte, err
		}
		return dataByte, nil
	}
	if isCompressionEnabled {
		dataByte, err = decompressData(dataByte)
		if err != nil {
			return dataByte, err
		}
		return dataByte, nil
	}
	return dataByte, err
}

func readDataFromNormalBucket(fdb *FDB, paths []string) (*gjson.Result, error) {
	data, err := fdb.readDataFile(paths)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	// if len(paths) > 0 && len(data.Array()) == 0 {
	// 	data, err = searchDataInPack(paths, fdb.DBPath, fdb.restoreFileFromPack)
	// 	if err != nil {
	// 		loggermdl.LogError(err)
	// 		return nil, err
	// 	}
	// }
	return data, nil
}

func getFileDataUsingInFileIndex(filePath string, bucket *Bucket, requestedFileType string, inFileIndexQuery []string) (string, error) {
	_, ok := bucket.InFileIndexMap[requestedFileType]
	if !ok {
		return "", errormdl.Wrap("infileIndex for specified fileType not found: " + requestedFileType)
	}
	if !filemdl.FileAvailabilityCheck(filePath) {
		return "", errormdl.Wrap("file not found at:" + filePath)
	}
	f, err := filemdl.Open(filePath)
	defer f.Close()
	if err != nil {
		loggermdl.LogError("err while opening file: ", filePath, err)
		return "", err
	}
	indexDataString, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", filePath, err)
		return "", err
	}
	indexData := gjson.Parse(indexDataString)
	indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQuery); i++ {
		if inFileIndexQuery[i] == GetAll {
			// return all results
			break
		}
		indexRows = indexRows.Get(inFileIndexQuery[i] + "#")
	}
	sb := strings.Builder{}
	sb.WriteString("[")
	indexRows.ForEach(func(key, indexRow gjson.Result) bool {
		// read files
		startOffSet := indexRow.Get("startOffset").Int()
		dataSize := indexRow.Get("dataSize").Int()
		if startOffSet == 0 || dataSize == 0 {
			return true
		}
		dataByte := []byte{'{', '}'}
		var err error
		// dataByte, err = filemdl.ReadFileFromOffset(f, startOffSet, dataSize)
		dataByte, err = getFileDataFromFile(f, startOffSet, dataSize)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		_, err = sb.Write(dataByte)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		sb.WriteString(",")

		return true // keep iterating
	})
	sb.WriteString("]")
	finalResult := strings.Replace(sb.String(), ",]", "]", 1)
	return finalResult, nil
}

func deletDataFromInfileIndex(filePath string, fileType string, infileIndexQueries []string) error {
	// get indexdata
	// get update index records
	// save updated index
	// update index size header
	// lock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(lock)
	// if err != nil {
	// 	return err
	// }

	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer f.Close()
	if err != nil {
		return err
	}

	indexDataStr, err := getInFileIndexData(f)
	if err != nil {
		return err
	}

	indexData := gjson.Parse(indexDataStr)
	indexRecordsToDelete := indexData.Get(`#[fileType==` + fileType + `]#`)
	// loggermdl.LogDebug("indexRecordsToDelete file type", indexRecordsToDelete)
	for _, query := range infileIndexQueries {
		indexRecordsToDelete = indexRecordsToDelete.Get(query + "#")
	}
	indexRowsToDelete := indexRecordsToDelete.Array()
	if len(indexRowsToDelete) == 0 {
		return ErrNoDataFound
	}
	updatedIndexRecords := indexData

	for _, indexRowToRemove := range indexRowsToDelete {
		updatedIndexRecords, err = removeIndexRow(updatedIndexRecords, indexRowToRemove.String())
		if err != nil {
			loggermdl.LogError("fail to delete record:", err)
			return errormdl.Wrap("fail to delete record:" + err.Error())
		}
	}

	footerOffset := getFooterOffset(f)
	if footerOffset == -1 {
		return errormdl.Wrap("fail to fetch infile index offset")
	}
	newIndexDataSize := len(updatedIndexRecords.String())
	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return err
	}
	err = setIndexDataInFile(f, footerOffset, updatedIndexRecords.String())
	if err != nil {
		loggermdl.LogError("fail to update infile index data :", err)
		return err
	}
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return err
	}
	return setFooterSize(f, int64(newIndexDataSize))
}

func removeIndexRow(indexRows gjson.Result, indexRowToDelete string) (gjson.Result, error) {

	indexRowObjs := indexRows.Array()
	if len(indexRowObjs) == 0 {
		return indexRows, errormdl.Wrap("no data found to update")
	}
	// loggermdl.LogDebug("indexRows", indexRows)

	foundIndexToDelete := -1
	for index, indexRowObj := range indexRowObjs {
		if indexRowObj.String() != "" && indexRowToDelete != "" && indexRowObj.String() == indexRowToDelete {
			foundIndexToDelete = index
			break
		}
	}

	if foundIndexToDelete == -1 {
		return indexRows, errormdl.Wrap("no record found to delete")
	}
	var err error
	updatedIndexDataString, err := sjson.Delete(indexRows.String(), strconv.Itoa(foundIndexToDelete))
	if err != nil {
		loggermdl.LogError(err)
	}
	return gjson.Parse(updatedIndexDataString), nil
}

// ReindexOnSpecialBucket - ReindexOnSpecialBucket
func ReindexOnSpecialBucket(fdbName string, indexID string) error {
	fdb, err := GetFDBInstance(fdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", fdbName)
		return errormdl.Wrap(("fdb instance not found: " + fdbName))
	}

	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: ", indexID)
		return errormdl.Wrap("index not found: " + indexID)
	}
	// find path to start file walk
	pathToStartWalk := fdb.DBPath
	for _, bucketID := range index.BucketSequence {
		bucket := fdb.buckets[bucketID]
		if bucket.IsDynamicName {
			break
		}
		pathToStartWalk = filepath.Join(pathToStartWalk, bucket.BucketNameQuery)
	}
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket := fdb.buckets[bucketID]
	indexDataMap := make(map[string]string)
	if !filemdl.FileAvailabilityCheck(pathToStartWalk) {
		return errormdl.Wrap("invalid path: " + pathToStartWalk)
	}

	ferr := filemdl.Walk(pathToStartWalk, func(filePath string, info os.FileInfo, err error) error {
		if err != nil {
			loggermdl.LogError(err)
			return nil
		}

		if !info.IsDir() {
			result, err := getFileDataUsingInFileIndex(filePath, bucket, FileTypeFDBIndex, []string{`#[fileType=="FDBIndex"]`})
			if err != nil {
				loggermdl.LogError(err)
				return nil
			}
			data := gjson.Parse(result)
			data = data.Get("0")
			pathToSave := strings.TrimPrefix(filePath, fdb.DBPath+string(filepath.Separator))
			valString, _ := sjson.Delete(data.String(), "fileType")
			// val= gjson.Parse(valString)
			indexDataMap[pathToSave] = strings.TrimSpace(valString)
		}

		return nil
	})
	if ferr != nil {
		return ferr
	}
	return reCreateFDBIndexFile(fdb, index, indexDataMap)
}

func reCreateFDBIndexFile(fdb *FDB, index *Index, indexDataMap map[string]string) error {
	// create or replace index
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	index.indexObj.Close()
	indexFilePath := filepath.Join(fdb.DBPath, INDEXFOLDER, index.IndexID)
	if filemdl.FileAvailabilityCheck(indexFilePath) {
		err := filemdl.DeleteFile(indexFilePath)
		if err != nil {
			return err
		}
	}
	db, err := fdb.GetIndexDB(index)
	if err != nil {
		return err
	}
	index.indexObj = db
	err = index.indexObj.ReplaceIndex(index.IndexID, "*", fns...)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// update index file by reading all data and updating index file
	err = index.indexObj.Update(func(tx *buntdb.Tx) error {

		for indexKey, indexData := range indexDataMap {
			_, _, err := tx.Set(indexKey, indexData, nil)
			if err != nil {
				loggermdl.LogError(err)
				return err
			}
		}

		return nil
	})
	return nil
}

// LogFDBIndexFile -LogFDBIndexFile
func LogFDBIndexFile(indexFilePath string, index *Index) error {
	// dbPath := filepath.Join(fdbPath, INDEXFOLDER)
	// loggermdl.LogDebug("in log fdb index")
	dataToStore := ``

	index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend(index.IndexID, func(key, value string) bool {
			dataToStore = dataToStore + key + IndexKeyValSeperator + value + lineBreak
			// rsJSON := gjson.Parse("[" + value + "]")
			return true
		})
	})

	dataByteToWrite := []byte(dataToStore)
	var hashError error
	dataByteToWriteRes, hashError := filemdl.ZipBytes(dataByteToWrite)
	if errormdl.CheckErr2(hashError) != nil {
		return errormdl.CheckErr2(hashError)
	}

	keyBytes, hashError := GetKeyWithFileNameAndDefaultKey(indexFilePath)
	if errormdl.CheckErr(hashError) != nil {
		return errormdl.CheckErr(hashError)
	}
	dataByteToWriteRes, hashError = securitymdl.AESEncrypt(dataByteToWriteRes, keyBytes)
	if errormdl.CheckErr1(hashError) != nil {
		return errormdl.CheckErr1(hashError)
	}

	return filemdl.WriteFile(indexFilePath, dataByteToWriteRes, true, false)
}

// GetKeyWithFileNameAndDefaultKey generates key using file name + Default key
func GetKeyWithFileNameAndDefaultKey(filePath string) ([]byte, error) {
	fileName := filepath.Base(filePath)
	fileNameBytes := []byte(fileName)
	fileNameBytes = append(fileNameBytes, defaultSecurityKey...)
	keyBytes, getHashError := hashmdl.Get128BitHash(fileNameBytes)
	if errormdl.CheckErr(getHashError) != nil {
		return nil, errormdl.CheckErr(getHashError)
	}
	return keyBytes[:], nil
}

// LoadFDBIndexFromFile -
func LoadFDBIndexFromFile(indexFilePath string, fdb *FDB, indexID string) error {
	index, found := fdb.GetFDBIndex(indexID)
	if !found {
		return errormdl.Wrap("index not found")
	}

	fileData, err := filemdl.FastReadFile(indexFilePath)
	if err != nil {
		loggermdl.LogError("failed to load FDB index from: ", indexFilePath)
		return err
	}
	// if isSecurityEnabled {
	keyBytes, hashError := GetKeyWithFileNameAndDefaultKey(indexFilePath)
	if errormdl.CheckErr(hashError) != nil {
		return errormdl.CheckErr(hashError)
	}
	fileData, err = securitymdl.AESDecrypt(fileData, keyBytes)
	if err != nil {
		loggermdl.LogError("failed to decrypt data", err)
		return err
	}
	// }
	fileData, err = filemdl.UnZipBytes(fileData)
	if err != nil {
		loggermdl.LogError("failed to decompress FDB index data")
		return err
	}
	data := string(fileData)
	indexRecords := strings.Split(data, lineBreak)
	indexDataMap := make(map[string]string)
	for _, indexRecord := range indexRecords {
		indexValues := strings.Split(indexRecord, IndexKeyValSeperator)
		if len(indexValues) == 2 {
			indexDataMap[indexValues[0]] = indexValues[1]
		}
	}
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}

	index.indexObj.Close()
	db, err := fdb.GetIndexDB(index)
	if err != nil {
		return err
	}
	index.indexObj = db
	err = index.indexObj.ReplaceIndex(index.IndexID, "*", fns...)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// update index file by reading all data and updating index file
	err = index.indexObj.Update(func(tx *buntdb.Tx) error {
		for key, val := range indexDataMap {
			tx.Set(key, val, nil)
		}
		return nil
	})
	return nil
}

func deleteDataFromNormalBucket(fdbPath string, index *Index, paths []string) (recordsDeleted int, errList []error) {
	index.indexObj.Update(func(tx *buntdb.Tx) error {
		for _, path := range paths {
			// delete index entry
			_, err := tx.Delete(path)
			if err != nil {
				errList = append(errList, errormdl.Wrap("unable to delete index :"+path+err.Error()))
				continue
			}
			path = filepath.Join(fdbPath, path)

			// dirPath, _ := filepath.Split(path)
			// search filepack
			// packPath, err := filepack.SearchPackFilePath(dirPath, fdbPath)
			// if err != nil && err.Error() != filepack.PackNotFound {
			// 	loggermdl.LogError("err", err)
			// 	continue
			// }
			// delete file from filepack
			// if err.Error() != filepack.PackNotFound {
			// err = filepack.DeleteFileFromPack(packPath, path)
			// if err != nil {
			// 	errList = append(errList, errormdl.Wrap("unable to delete file from Pack :"+path))
			// 	continue
			// }
			// }
			// delete file

			err = deleteNormalFile(path)

			if err != nil {
				errList = append(errList, errormdl.Wrap("unable to delete file : "+path+err.Error()))
				continue
			}
			recordsDeleted++
		}
		return nil
	})

	// TODO: Currently index data is overwritten by new data.
	err := UpdateIndexLazyObjectInCache(index.IndexID, index)
	if err != nil {
		loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
		return 0, []error{errormdl.Wrap("failed to update index data in lazy writer cache")}
	}
	return
}

func deleteNormalFile(filePath string) error {
	// lock, ferr := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(lock)

	// if ferr != nil {
	// 	return errormdl.Wrap("unable to delete file : " + filePath + ferr.Error())
	// }
	return filemdl.DeleteFile(filePath)
}
func deleteDataFromInFileIndexBucket(fdbPath string, bucket *Bucket, paths []string, rs *gjson.Result, infileIndexQueries []string) (recordsDeleted int, errList []error) {

	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		// loggermdl.LogError("fileType value not provided")
		return recordsDeleted, []error{errormdl.Wrap("please specify fileType")}
	}

	_, ok := bucket.InFileIndexMap[fileType]
	if !ok {
		return recordsDeleted, []error{errormdl.Wrap("infileIndex for specified fileType not found")}
	}
	noDataFoundCnt := 0
	for _, filePath := range paths {
		path := filepath.Join(fdbPath, filePath)

		err := deletDataFromInfileIndex(path, fileType, infileIndexQueries)
		if err == ErrNoDataFound {
			noDataFoundCnt++
			continue
		}
		if err != nil {
			errList = append(errList, errormdl.Wrap("fail to delete records from: "+path+" : "+err.Error()))
		}
		recordsDeleted++
	}
	if noDataFoundCnt == len(paths) {
		errList = []error{ErrNoDataFound}
	}
	return recordsDeleted, errList
}

// SaveDataInFDB -SaveDataInFDB
func SaveDataInFDB(dbName, indexID string, rs *gjson.Result) error {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return errormdl.Wrap("INDEX not found: " + indexID)
	}
	path, err := fdb.resolveIndex(index, rs)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}

	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return errormdl.Wrap("Bucket not found: " + bucketID)
	}
	filePath := filepath.Join(fdb.DBPath, path)

	// check is infile index present
	// if yes
	switch bucket.BucketType {
	case BucketTypeSimple:
		err := saveDataInNormalBucket(filePath, rs)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return err
		}
	case BucketTypePack:
		err := saveDataInPackBucket(fdb, filePath, index, rs)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return errormdl.CheckErr(err)
		}
	case BucketTypeAppend:
		err := saveDataInAppendBucket(filePath, rs, bucket)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return errormdl.CheckErr(err)
		}

	default:
		return errormdl.Wrap("Please provide valid bucket type")
	}
	// loggermdl.LogDebug("index", *index)
	json, err := createIndexJSON(index, rs)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	// path = path + ".json"
	err = index.indexObj.Update(func(tx *buntdb.Tx) error {
		// optimization : if data not changed then dont update
		_, _, err := tx.Set(path, json, nil)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		return nil
	})
	// TODO: Currently index data is overwritten by new data.
	err = UpdateIndexLazyObjectInCache(indexID, index)
	if err != nil {
		loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
		return errormdl.Wrap("failed to update index data in lazy writer cache")
	}
	return nil
}

// ReadDataFromFDB - return records with matching query
func ReadDataFromFDB(dbName, indexID string, rs *gjson.Result, query []string, infileIndexQuery []string) (*gjson.Result, error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return nil, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return nil, errormdl.Wrap("INDEX not found: " + indexID)
	}
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return nil, errormdl.Wrap("Bucket not found: " + bucketID)
	}
	filePaths := []string{}
	// loggermdl.LogDebug(query)
	index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend(index.IndexID, func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(query); i++ {
				loggermdl.LogError(query[i])
				if query[i] == GetAll {
					// return all results
					break
				}
				rsJSON = rsJSON.Get(query[i] + "#")
			}
			if rsJSON.Value() != nil {
				filePaths = append(filePaths, key)
			}
			return true
		})
	})
	resultToReturn := gjson.Result{}
	if len(filePaths) == 0 {
		loggermdl.LogError("files not found")
		return &resultToReturn, nil
	}
	resultArray := "[]"
	if bucket.BucketType == BucketTypePack {
		for _, path := range filePaths {
			filePath := filepath.Join(fdb.DBPath, path)
			requestedFileType := rs.Get("fileType").String()
			if len(requestedFileType) == 0 {
				return &resultToReturn, errormdl.Wrap("please specify fileType")
			}
			bucketID := index.BucketSequence[len(index.BucketSequence)-1]
			bucket := fdb.buckets[bucketID]
			result, err := getFileDataUsingInFileIndex(filePath, bucket, requestedFileType, infileIndexQuery)
			if err != nil {
				loggermdl.LogError(err)
				return &resultToReturn, err
			}
			// loggermdl.LogDebug("result", result)
			for _, val := range gjson.Parse(result).Array() {
				resultArray, _ = sjson.Set(resultArray, "-1", val.Value())
			}
		}
		resultToReturn = gjson.Parse(resultArray)
		return &resultToReturn, nil
	} else if bucket.BucketType == BucketTypeSimple {
		return readDataFromNormalBucket(fdb, filePaths)
	} else {
		return nil, errormdl.Wrap("Operation not available on bucket type " + bucket.BucketType)
	}
	// loggermdl.LogError("resultToReturn - ", resultToReturn)

}

// UpdateDataInFDB - UpdateDataInFDB
func UpdateDataInFDB(dbName, indexID string, rs *gjson.Result, query []string, infileIndexQuery []string) (*gjson.Result, []error) {
	var errList []error
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return nil, errList
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return nil, []error{errormdl.Wrap("INDEX not found: " + indexID)}
	}
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return nil, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
	}

	if bucket.BucketType == BucketTypePack {
		requestedFileType := rs.Get("fileType").String()
		if len(requestedFileType) == 0 {
			loggermdl.LogError("file type value not provided")
			return nil, []error{errormdl.Wrap("please specify fileType")}
		}
	}
	filePaths := []string{}
	indexKeyValMap := make(map[string]string)
	// loggermdl.LogDebug(query)
	index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend(index.IndexID, func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(query); i++ {
				if query[i] == GetAll {
					// return all results
					break
				}
				rsJSON = rsJSON.Get(query[i] + "#")
			}
			if rsJSON.Value() != nil {
				filePaths = append(filePaths, key)
				indexKeyValMap[key] = rsJSON.Array()[0].String()
			}
			return true
		})
	})

	// loggermdl.LogDebug("filePaths", filePaths)
	resultToReturn := gjson.Result{}

	if len(filePaths) == 0 {
		loggermdl.LogError("files not found")
		return nil, []error{ErrNoDataFound}
	}
	result := &gjson.Result{}
	var dataUpdatedAtFilePaths []string
	switch bucket.BucketType {
	case BucketTypeSimple:
		result, dataUpdatedAtFilePaths, errList = updateDataInNormalBucket(fdb, filePaths, rs)
	case BucketTypePack:
		result, dataUpdatedAtFilePaths, errList = updateDataInFileIndexBucket(fdb, bucket, filePaths, rs, infileIndexQuery)
	default:
		loggermdl.LogError("invalid bucket type")
		return nil, []error{errormdl.Wrap("invalid bucket type - " + bucket.BucketType)}
	}

	if len(dataUpdatedAtFilePaths) == 0 {
		loggermdl.LogError("no data found to update")
		return &resultToReturn, []error{ErrNoDataFound}
	}
	// path = path + ".json"
	index.indexObj.Update(func(tx *buntdb.Tx) error {
		// optimization : if data not changed then dont update
		for _, path := range dataUpdatedAtFilePaths {
			json := indexKeyValMap[path]
			updatedJSON, err := updateIndexJSON(index, json, rs)
			if err != nil {
				errList = append(errList, err)
			}
			_, _, err = tx.Set(path, updatedJSON, nil)
			if err != nil {
				errList = append(errList, err)
			}
		}
		return nil
	})

	// TODO: Currently index data is overwritten by new data.
	err = UpdateIndexLazyObjectInCache(indexID, index)
	if err != nil {
		loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
		return nil, []error{errormdl.Wrap("failed to update index data in lazy writer cache")}
	}

	return result, errList
}

// DeleteDataFromFDB -
func DeleteDataFromFDB(dbName string, indexID string, rs *gjson.Result, queries []string, infileIndexQueries []string) (recordsDeleted int, errList []error) {
	// errList := make([]error, 0)

	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found for: ", dbName)
		return recordsDeleted, []error{err}
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: " + indexID)
		return recordsDeleted, []error{errormdl.Wrap("index not found: " + indexID)}
	}
	filePaths := []string{}
	// loggermdl.LogDebug(query)
	index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend(index.IndexID, func(key, value string) bool {
			// loggermdl.LogDebug(key, value)
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(queries); i++ {
				rsJSON = rsJSON.Get(queries[i] + "#")
			}
			if rsJSON.Value() != nil {
				filePaths = append(filePaths, key)
			}
			return true
		})
	})
	if len(filePaths) == 0 {
		loggermdl.LogError("no data found to delete")
		return recordsDeleted, []error{ErrNoDataFound}
	}
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return recordsDeleted, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
	}

	if bucket.BucketType == BucketTypeSimple {
		recordsDeleted, errList = deleteDataFromNormalBucket(fdb.DBPath, index, filePaths)
	} else if bucket.BucketType == BucketTypePack {
		recordsDeleted, errList = deleteDataFromInFileIndexBucket(fdb.DBPath, bucket, filePaths, rs, infileIndexQueries)
	} else {
		return recordsDeleted, []error{errormdl.Wrap("Operation not allowed on bucket type: " + bucket.BucketType)}
	}
	return recordsDeleted, errList
}

// GetBucketByIndexID - return bucket by specified indexID
func GetBucketByIndexID(dbName, indexID string) (Bucket, error) {
	fdb, err := GetFDBInstance(dbName)
	bucket := Bucket{}
	if err != nil {
		loggermdl.LogError("fdb instance not found for: ", dbName)
		return bucket, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: " + indexID)
		return bucket, errormdl.Wrap("index not found: " + indexID)
	}

	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucketPtr, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return bucket, errormdl.Wrap("Bucket not found: " + bucketID)
	}
	return *bucketPtr, nil
}
