package corefdb

import (
	"log"
	"path/filepath"
	"testing"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"github.com/tidwall/buntdb"
	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
)

var dbInstance *FDB
var bucketInstance *Bucket
var indexInstance *Index

func init() {
	Init(true, false, false, "")

	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}

	dbInstance = db
	// step 1:  create bucket
	bucket := db.GetNewBucket("Candidates", false, &Bucket{})
	bucketInstance = bucket
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(bucket)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
		IndexField{
			FieldName: "class",
			Query:     "class",
		},
	}
	i.SetFields(fields...)
	indexInstance = i
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}
}

func TestReindex(t *testing.T) {

	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("Candidates", false, &Bucket{})
	b.SetBucketType(BucketTypeSimple)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
		IndexField{
			FieldName: "class",
			Query:     "class",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	data, _ := sjson.Set("", "name", "vivek")
	data, _ = sjson.Set(data, "studentId", 1000)

	data, _ = sjson.Set(data, "class", "TY_MCA")
	data, _ = sjson.Set(data, "age", 26)

	studentObj := gjson.Parse(data)
	// step 3: save data
	err = SaveDataInFDB("myfdb", i.IndexID, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	// step 4: fetch data
	result, err := ReadDataFromFDB("myfdb", i.IndexID, &studentObj, []string{`#[name=="vivek"]`}, []string{""})
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogDebug("before reindex", result)

	// step 5: delete index file
	err = i.indexObj.Close()
	if err != nil {
		log.Fatal(err)
	}

	err = filemdl.DeleteFile("/home/vivekn/fdb_data/index/studentId")
	if err != nil {
		loggermdl.LogError(err)
		log.Fatal(err)
	}

	// step 6: reindex
	err = db.ReIndex(i.IndexID)
	if err != nil {
		log.Fatal(err)
	}

	// step 7: fetch data
	data2, err := ReadDataFromFDB("myfdb", i.IndexID, &studentObj, []string{`#[name=="vivek"]`}, []string{""})

	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogDebug("after reindex", data2)
}

func BenchmarkReindex(b *testing.B) {

	for i := 0; i < 300; i++ {
		data, _ := sjson.Set(`{"age":26,"class":"SY_MCA","studentId":10008,"name":"vivek"}`, "studentId", 1000+i)
		obj := gjson.Parse(data)
		SaveDataInFDB(dbInstance.DBName, indexInstance.IndexID, &obj)
	}
	for index := 0; index < b.N; index++ {
		dbInstance.ReIndex(indexInstance.IndexID)
	}
}

func TestSaveDataInInfileBucket(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypePack)

	infileIndexProfile := InFileIndex{
		FileType: "Profile",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "class",
				Query:     "class",
			},
		},
	}
	infileIndexExam := InFileIndex{
		FileType: "Exam",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "examId",
				Query:     "examId",
			},
		},
	}
	// loggermdl.LogDebug(infileIndexExam)
	b.SetInFileIndex(infileIndexProfile)
	b.SetInFileIndex(infileIndexExam)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	// data, _ := sjson.Set("", "name", "ajay")
	// data, _ = sjson.Set(data, "studentId", 10013)
	// data, _ = sjson.Set(data, "class", "TY_MCA")
	// data, _ = sjson.Set(data, "age", 24)
	// data, _ = sjson.Set(data, "fileType", "Profile")

	data, _ := sjson.Set("", "name", "ajay")
	data, _ = sjson.Set(data, "studentId", 1234)
	data, _ = sjson.Set(data, "examName", "unit2")
	data, _ = sjson.Set(data, "totalQuestion", 50)
	data, _ = sjson.Set(data, "marks", 26)
	data, _ = sjson.Set(data, "examId", "MATH001")
	data, _ = sjson.Set(data, "fileType", "Exam")

	//////// ------------vijay

	// data, _ := sjson.Set("", "name", "vijay")
	// data, _ = sjson.Set(data, "studentId", 10014)
	// data, _ = sjson.Set(data, "class", "SY_MCA")
	// data, _ = sjson.Set(data, "age", 23)
	// data, _ = sjson.Set(data, "fileType", "Profile")

	// data, _ := sjson.Set("", "studentId", 10014)
	// data, _ = sjson.Set(data, "name", "vijay")
	// data, _ = sjson.Set(data, "examName", "Unit3")
	// data, _ = sjson.Set(data, "totalQuestion", 100)
	// data, _ = sjson.Set(data, "marks", 32)
	// data, _ = sjson.Set(data, "fileType", "Exam")

	// studentList := []gjson.Result{}
	// for index := 0; index < 5; index++ {
	// 	data, _ = sjson.Set(data, "studentId", index+1)
	// 	studentList = append(studentList, gjson.Parse(data))
	// }
	// for _, studentObj := range studentList {
	// 	err = SaveDataInFDB("myfdb", i.IndexID, &studentObj)
	// 	if err != nil {
	// 		log.Fatal(err)
	// 	}
	// }
	studentObj := gjson.Parse(data)

	err = SaveDataInFDB("myfdb", i.IndexID, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LogFDBIndexFile(indexFilePath, i)
	if err != nil {
		log.Fatal(err)
	}
}

func TestSaveDataInNormalBucket(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypeSimple)

	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	// data, _ := sjson.Set("", "name", "ajay")
	// data, _ = sjson.Set(data, "studentId", 10013)
	// data, _ = sjson.Set(data, "class", "TY_MCA")
	// data, _ = sjson.Set(data, "age", 24)
	// data, _ = sjson.Set(data, "fileType", "Profile")

	data, _ := sjson.Set("", "name", "ajay")
	data, _ = sjson.Set(data, "studentId", 1234)
	data, _ = sjson.Set(data, "examName", "unit2")
	data, _ = sjson.Set(data, "totalQuestion", 50)
	data, _ = sjson.Set(data, "marks", 26)
	data, _ = sjson.Set(data, "examId", "MATH001")
	data, _ = sjson.Set(data, "fileType", "Exam")

	//////// ------------vijay

	// data, _ := sjson.Set("", "name", "vijay")
	// data, _ = sjson.Set(data, "studentId", 10014)
	// data, _ = sjson.Set(data, "class", "SY_MCA")
	// data, _ = sjson.Set(data, "age", 23)
	// data, _ = sjson.Set(data, "fileType", "Profile")

	// data, _ := sjson.Set("", "studentId", 10014)
	// data, _ = sjson.Set(data, "name", "vijay")
	// data, _ = sjson.Set(data, "examName", "Unit3")
	// data, _ = sjson.Set(data, "totalQuestion", 100)
	// data, _ = sjson.Set(data, "marks", 32)
	// data, _ = sjson.Set(data, "fileType", "Exam")

	// studentList := []gjson.Result{}
	// for index := 0; index < 5; index++ {
	// 	data, _ = sjson.Set(data, "studentId", index+1)
	// 	studentList = append(studentList, gjson.Parse(data))
	// }
	// for _, studentObj := range studentList {
	// 	err = SaveDataInFDB("myfdb", i.IndexID, &studentObj)
	// 	if err != nil {
	// 		log.Fatal(err)
	// 	}
	// }
	studentObj := gjson.Parse(data)

	err = SaveDataInFDB("myfdb", i.IndexID, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LogFDBIndexFile(indexFilePath, i)
	if err != nil {
		log.Fatal(err)
	}
}
func TestGetDataFromInFileBucket(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypePack)

	infileIndexProfile := InFileIndex{
		FileType: "Profile",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "class",
				Query:     "class",
			},
		},
	}
	infileIndexExam := InFileIndex{
		FileType: "Exam",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "examId",
				Query:     "examId",
			},
		},
	}
	inFileFDBIndex := InFileIndex{
		FileType: "FDBIndex",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "name",
				Query:     "name",
			},
		},
	}
	// loggermdl.LogDebug(infileIndexExam)
	b.SetInFileIndex(infileIndexProfile)
	b.SetInFileIndex(infileIndexExam)
	b.SetInFileIndex(inFileFDBIndex)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	//   create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LoadFDBIndexFromFile(indexFilePath, db, i.IndexID)
	if err != nil {
		log.Fatal(err)
	}
	// queries := []string{`#[name==ajay]`}

	data, _ := sjson.Set("", "fileType", "Exam")
	inFileIndexQueries := []string{`#[examId=="MATH001"]`}

	// data, _ := sjson.Set("", "fileType", "Profile")
	// inFileIndexQueries := []string{`#[class==TY_MCA]`}

	queries := []string{`#[name=="ajay"]`}
	// inFileIndexQueries := []string{`#[class==SY_MCA]`}

	studentObj := gjson.Parse(data)
	// step 3: save data
	result, err := ReadDataFromFDB("myfdb", i.IndexID, &studentObj, queries, inFileIndexQueries)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogDebug("result", result)
}

func TestUpdateData(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypePack)

	infileIndexProfile := InFileIndex{
		FileType: "Profile",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "class",
				Query:     "class",
			},
		},
	}
	infileIndexExam := InFileIndex{
		FileType: "Exam",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "marks",
				Query:     "marks",
			},
		},
	}
	// loggermdl.LogDebug(infileIndexExam)
	b.SetInFileIndex(infileIndexProfile)
	b.SetInFileIndex(infileIndexExam)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}
	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LoadFDBIndexFromFile(indexFilePath, db, i.IndexID)
	if err != nil {
		log.Fatal(err)
	}
	// data, _ := sjson.Set("", "name", "ajay")
	// data, _ = sjson.Set(data, "studentId", 10013)
	// data, _ = sjson.Set(data, "class", "TY_MCA")
	// data, _ = sjson.Set(data, "age", 24)
	// data, _ = sjson.Set(data, "fileType", "Profile")

	// data, _ := sjson.Set("", "studentId", 10013)
	// data, _ = sjson.Set(data, "name", "ajay")
	// data, _ = sjson.Set(data, "examName", "Unit1")
	// data, _ = sjson.Set(data, "totalQuestion", 50)
	// data, _ = sjson.Set(data, "marks", 30)
	// data, _ = sjson.Set(data, "fileType", "Exam")

	//////// ------------ajay

	// data, _ := sjson.Set("", "name", "vijay")
	// data, _ = sjson.Set(data, "studentId", 10014)
	// data, _ = sjson.Set(data, "class", "SY_MCA")
	// data, _ = sjson.Set(data, "age", 23)
	// data, _ = sjson.Set(data, "fileType", "Profile")

	data, _ := sjson.Set("", "studentId", 10014)
	// data, _ = sjson.Set(data, "name", "vijay")
	// data, _ = sjson.Set(data, "examName", "t")
	data, _ = sjson.Set(data, "totalQuestion", 40)
	data, _ = sjson.Set(data, "marks", 32)
	data, _ = sjson.Set(data, "fileType", "Exam")

	queries := []string{`#[name=="vijay"]`}
	infileIndexQueries := []string{`#[marks==32]`}
	studentObj := gjson.Parse(data)
	// step 3: save data
	updatedData, errList := UpdateDataInFDB("myfdb", i.IndexID, &studentObj, queries, infileIndexQueries)
	if len(errList) > 0 {
		loggermdl.LogError(errList)
	}
	loggermdl.LogDebug("updatedData", updatedData)
}

func TestReindexOnSpecialBucket(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypePack)
	infileIndexFDB := InFileIndex{
		FileType: "FDBIndex",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "name",
				Query:     "name",
			},
		},
	}
	// loggermdl.LogDebug(infileIndexExam)
	b.SetInFileIndex(infileIndexFDB)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = ReindexOnSpecialBucket("myfdb", i.IndexID)
	if err != nil {
		log.Fatal(err)
	}

}

func TestLogFDBIndexFile(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypePack)

	infileIndexProfile := InFileIndex{
		FileType: "Profile",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "class",
				Query:     "class",
			},
		},
	}
	infileIndexExam := InFileIndex{
		FileType: "Exam",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "marks",
				Query:     "marks",
			},
		},
	}
	// loggermdl.LogDebug(infileIndexExam)
	b.SetInFileIndex(infileIndexProfile)
	b.SetInFileIndex(infileIndexExam)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	data, _ := sjson.Set("", "name", "ajay")
	data, _ = sjson.Set(data, "examName", "unit2")
	data, _ = sjson.Set(data, "totalQuestion", 50)
	data, _ = sjson.Set(data, "marks", 26)
	data, _ = sjson.Set(data, "fileType", "Exam")

	studentList := []gjson.Result{}
	for index := 0; index < 1; index++ {
		data, _ = sjson.Set(data, "studentId", index+1)
		studentList = append(studentList, gjson.Parse(data))
	}
	// step 3: save data
	for _, studentObj := range studentList {
		// data, _ = sjson.Set(data, "studentId", v+1)
		// studentObj := gjson.Parse(data)
		err = SaveDataInFDB("myfdb", i.IndexID, &studentObj)
		if err != nil {
			log.Fatal(err)
		}
	}
	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LogFDBIndexFile(indexFilePath, i)
	if err != nil {
		log.Fatal(err)
	}

}

func TestLoadFDBIndexFromFile(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypePack)

	infileIndexProfile := InFileIndex{
		FileType: "Profile",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "class",
				Query:     "class",
			},
		},
	}
	infileIndexExam := InFileIndex{
		FileType: "Exam",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "marks",
				Query:     "marks",
			},
		},
	}
	// loggermdl.LogDebug(infileIndexExam)
	b.SetInFileIndex(infileIndexProfile)
	b.SetInFileIndex(infileIndexExam)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LoadFDBIndexFromFile(indexFilePath, db, i.IndexID)
	if err != nil {
		log.Fatal(err)
	}

	data, _ := sjson.Set("", "fileType", "Exam")
	inFileIndexQueries := []string{`#[marks==26]`}

	// data, _ := sjson.Set("", "fileType", "Profile")
	// inFileIndexQueries := []string{`#[class==TY_MCA]`}

	queries := []string{`#[name=="ajay"]`}
	// inFileIndexQueries := []string{`#[class==SY_MCA]`}

	studentObj := gjson.Parse(data)
	// step 3: save data
	result, err := ReadDataFromFDB("myfdb", i.IndexID, &studentObj, queries, inFileIndexQueries)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogDebug("result", result)
}

func TestSaveDataInAppendBucket(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("appendBucket", false, &Bucket{})
	b.SetBucketType(BucketTypeAppend)

	// loggermdl.LogDebug(infileIndexExam)
	i, err := db.GetNewIndex("studentProfId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	data, _ := sjson.Set("", "name", "ajay")
	data, _ = sjson.Set(data, "studentProfId", "10013")
	data, _ = sjson.Set(data, "class", "TY_MCA")
	data, _ = sjson.Set(data, "age", 24)
	data, _ = sjson.Set(data, "fileType", "Profile")
	studentProfObj := gjson.Parse(data)

	// data2, _ := sjson.Set("", "name", "ajay")
	// data2, _ = sjson.Set(data2, "studentId", 10013)
	// data2, _ = sjson.Set(data2, "examName", "unit2")
	// data2, _ = sjson.Set(data2, "totalQuestion", 50)
	// data2, _ = sjson.Set(data2, "marks", 26)
	// data2, _ = sjson.Set(data2, "fileType", "Exam")
	// studentExamObj := gjson.Parse(data2)
	err = SaveDataInFDB("myfdb", i.IndexID, &studentProfObj)
	if err != nil {
		log.Fatal(err)
	}
	// err = SaveDataInAppendBucket("myfdb", i.IndexID, &studentExamObj)
	// if err != nil {
	// 	log.Fatal(err)
	// }
	// indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)
	// err = LogFDBIndexFile(indexFilePath, i)
	// if err != nil {
	// 	log.Fatal(err)
	// }
}

func TestLoadFDBEncryptedData(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("appendBucket", false, &Bucket{})
	b.SetBucketType(BucketTypePack)

	infileIndexProfile := InFileIndex{
		FileType: "Profile",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "class",
				Query:     "class",
			},
		},
	}
	infileIndexExam := InFileIndex{
		FileType: "Exam",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "marks",
				Query:     "marks",
			},
		},
	}
	// loggermdl.LogDebug(infileIndexExam)
	b.SetInFileIndex(infileIndexProfile)
	b.SetInFileIndex(infileIndexExam)
	i, err := db.GetNewIndex("studentProfId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}

	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LoadFDBIndexFromFile(indexFilePath, db, i.IndexID)
	if err != nil {
		log.Fatal(err)
	}

	i.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend(i.IndexID, func(key, value string) bool {
			loggermdl.LogDebug("key ", key, "value", value)
			return true
		})
	})
}

func TestDeleteDataFromInfileBucket(t *testing.T) {
	// TestSaveDataInFDB(t)

	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypeSimple)

	infileIndexProfile := InFileIndex{
		FileType: "Profile",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "class",
				Query:     "class",
			},
		},
	}
	infileIndexExam := InFileIndex{
		FileType: "Exam",
		IndexFields: []IndexField{
			IndexField{
				FieldName: "examId",
				Query:     "examId",
			},
		},
	}
	b.SetInFileIndex(infileIndexProfile)
	b.SetInFileIndex(infileIndexExam)
	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}
	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LoadFDBIndexFromFile(indexFilePath, db, i.IndexID)
	if err != nil {
		log.Fatal(err)
	}
	// data, _ := sjson.Set("", "fileType", "Exam")
	// data, _ = sjson.Set(data, "studentId", 1234)

	studentObj := gjson.Result{}

	// err = SaveDataInFDB("myfdb", i.IndexID, &studentObj)
	// if err != nil {
	// 	log.Fatal(err)
	// }

	queries := []string{`#[name=="ajay"]`}
	infileIndexQueries := []string{`#[examId==MATH001]`}
	// studentObj := gjson.Parse(data)
	recordsDeleted, errList := DeleteDataFromFDB("myfdb", i.IndexID, &studentObj, queries, infileIndexQueries)
	if len(errList) > 0 {
		loggermdl.LogError(errList)
	}
	loggermdl.LogDebug("recordsDeleted", recordsDeleted)
}

func TestDeleteDataFromNormalBucket(t *testing.T) {
	db, err := CreateFDBInstance("/home/vivekn/fdb_data", "myfdb", false)
	if err != nil {
		log.Fatal("CreateFDBInstance = ", err)
	}
	// step 1:  create bucket
	b := db.GetNewBucket("bucketPackType", false, &Bucket{})
	b.SetBucketType(BucketTypeSimple)

	i, err := db.GetNewIndex("studentId", true)
	if err != nil {
		log.Fatal(err)
	}
	i.SetBucket(b)
	fields := []IndexField{
		IndexField{
			FieldName: "name",
			Query:     "name",
		},
	}
	i.SetFields(fields...)
	// step 2:  create index
	err = db.CreateIndex(i)
	if err != nil {
		log.Fatal(err)
	}
	indexFilePath := filepath.Join(db.DBPath, INDEXFOLDER, i.IndexNameQuery)

	err = LoadFDBIndexFromFile(indexFilePath, db, i.IndexID)
	if err != nil {
		log.Fatal(err)
	}
	queries := []string{`#[name=="ajay"]`}
	infileIndexQueries := []string{`#[examId==MATH001]`}

	// data, _ := sjson.Set("", "fileType", "Exam")
	// data, _ := sjson.Set(data, "studentId", 1234)

	studentObj := gjson.Result{}

	recordsDeletedCnt, errList := DeleteDataFromFDB("myfdb", i.IndexID, &studentObj, queries, infileIndexQueries)
	if len(errList) > 0 {
		loggermdl.LogError(errList)
	}
	loggermdl.LogDebug("recordsDeletedCnt", recordsDeletedCnt)

}
