package filepack

import (
	"fmt"
	"os"
	"path/filepath"
	"strings"
	"sync"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/hashmdl"

	"github.com/tidwall/gjson"

	"github.com/tidwall/sjson"

	"strconv"
)

const (
	// PackNotFound - PackNotFound
	PackNotFound = "packNotFound"
)

var (
	encryptionKey = "3Kc7$3cu31tyCFDB"
)

// Init - Init
func Init(key string) {
	encryptionKey = key
}

// CreatePackFile - creating new/initial pack file
func CreatePackFile(filePathArray []string, packCreationPath string) error {
	// open pack file to read
	f, err := filemdl.OpenFile(packCreationPath,
		os.O_CREATE|os.O_WRONLY, 0644)
	if err != nil {
		return err
	}
	defer f.Close()

	f.WriteAt([]byte("HHHHHHHHHHHHHHHHHHHHHHHHHHHHHH"), 0)
	indexTableStartOffset := int64(30)

	indexTableRecords := "[]"
	for index := 0; index < len(filePathArray); index++ {
		// read data of given file
		newFileData, readErr := filemdl.FastReadFile(filePathArray[index])
		if readErr != nil {
			return readErr
		}

		newFileSize := int64(len(newFileData))

		// create hash for file content
		hashText, err := getHashOfFile(newFileData)
		if err != nil {
			loggermdl.LogError("failed to create pack: ", err)
			return err
		}

		// write content to file
		f.WriteAt((newFileData), indexTableStartOffset)

		// append new record to indext table
		indexTableRecords = createNewIndexRecord(indexTableRecords, filePathArray[index], indexTableStartOffset, newFileSize, hashText)

		// increment startOffset with file size
		indexTableStartOffset += newFileSize
	}

	// write final updated index table to pack file
	return wrapUpSession(indexTableStartOffset, []byte(indexTableRecords), packCreationPath)
}

// CreateUpdatedPack - This create an new pack by checking updated contents
func CreateUpdatedPack(filePathArray []string, packFilePath string, destinationPackFilePath string) error {
	// fetch indexTableRecords and it's startOffset from pack file
	indexTableRecords, _, err := fetchIndexTable(packFilePath)
	if err != nil {
		return err
	}

	listOfUpdatedFiles := []string{}

	for index := 0; index < len(filePathArray); index++ {
		// read data of given file
		newFileData, readErr := filemdl.FastReadFile(filePathArray[index])
		if readErr != nil {
			loggermdl.LogError("failed to create pack: ", readErr)
			return readErr
		}

		// calculate hash of source file content
		newHashText, err := getHashOfFile(newFileData)
		if err != nil {
			loggermdl.LogError("failed to create pack: ", err)
			return err
		}

		// fetch hashText of given file in pack file
		existingFileHashText := getFileHashTextFromPack(indexTableRecords, filePathArray[index])

		// compare is file hash with existing record in pack file
		if newHashText != existingFileHashText {
			listOfUpdatedFiles = append(listOfUpdatedFiles, filePathArray[index])
		}
	}

	if len(listOfUpdatedFiles) == 0 {
		// no changes to create pack
		return nil
	}

	// create new pack with only changed content
	return CreatePackFile(listOfUpdatedFiles, destinationPackFilePath)
}

// MergePack - MergePack
func MergePack(sourcePackFilePath string, destinationPackFilePath string) error {
	// open destination pack file to read
	f, err := filemdl.OpenFile(destinationPackFilePath,
		os.O_CREATE|os.O_WRONLY, 0644)
	if err != nil {
		return err
	}
	defer f.Close()

	// fetch records from sourcePackFilePath
	sourcePackIndexTableRecords, _, err := fetchIndexTable(sourcePackFilePath)
	if err != nil {
		return err
	}

	//fetch All Filenames from sourcePackFilePath
	fileListFromSourcePack := gjson.Get(sourcePackIndexTableRecords, "#.fileName").Array()

	// fetch indexTableRecords and it's startOffset from pack file
	destinationPackIndexTableRecords, destinationPackIndexTableStartOffset, err := fetchIndexTable(destinationPackFilePath)
	if err != nil {
		return err
	}

	for index := 0; index < len(fileListFromSourcePack); index++ {
		// read data of given file
		newFileData, err := GetFileDataFromPack(sourcePackFilePath, fileListFromSourcePack[index].String())
		if err != nil {
			loggermdl.LogError("failed to merge pack: ", err)
			return err
		}
		// calculate length of source file content
		newFileSize := int64(len(newFileData))

		// create hash for file content
		hashText, err := getHashOfFile(newFileData)
		if err != nil {
			loggermdl.LogError("failed to merge pack: ", err)
			return err
		}
		// fetch existing file size from pack file
		existingFileSize := getFileSizeFromPack(destinationPackIndexTableRecords, fileListFromSourcePack[index].String())

		// fetch startOffset of given file in pack file if exist (If not then it is new file)
		existingFileStartOffset := getFileStartOffsetFromPack(destinationPackIndexTableRecords, fileListFromSourcePack[index].String())

		// compare is file smaller or larger than existing record in pack file
		if newFileSize <= existingFileSize {
			// write changed content to file
			f.WriteAt(newFileData, existingFileStartOffset)

			// update indext table with updated changes
			destinationPackIndexTableRecords = updateJSON(destinationPackIndexTableRecords, fileListFromSourcePack[index].String(), existingFileStartOffset, newFileSize, hashText)
		} else {
			// write changed content to file
			f.WriteAt(newFileData, destinationPackIndexTableStartOffset)

			// check it is existing file or new file
			if existingFileStartOffset == 0 {
				// append new record to indext table
				destinationPackIndexTableRecords = createNewIndexRecord(destinationPackIndexTableRecords, fileListFromSourcePack[index].String(), destinationPackIndexTableStartOffset, newFileSize, hashText)
			} else {
				// update indext table with updated changes
				destinationPackIndexTableRecords = updateJSON(destinationPackIndexTableRecords, fileListFromSourcePack[index].String(), destinationPackIndexTableStartOffset, newFileSize, hashText)
			}
			// increment startOffset for moving index table ahead
			destinationPackIndexTableStartOffset += newFileSize
		}
	}

	// write final updated index table to pack file
	return wrapUpSession(destinationPackIndexTableStartOffset, []byte(destinationPackIndexTableRecords), destinationPackFilePath)
}

// InsertFiles - inserts file data in pack
func InsertFiles(packPath string, filePathArray []string) error {
	mutex := sync.Mutex{}
	mutex.Lock()
	f, err := filemdl.OpenFile(packPath, os.O_RDWR, 0777)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	defer f.Close()
	indexTableRecords, indexStartOffset, err := fetchIndexTableMeta(f)
	if err != nil {
		loggermdl.LogError("failed to insert file in pack: ", err)
		return err
	}
	if len(indexTableRecords) == 0 {
		indexTableRecords = "[]"
	}

	for _, filePath := range filePathArray {

		// read data of given file
		newFileData, readErr := filemdl.FastReadFile(filePath)
		if readErr != nil {
			return readErr
		}

		newFileSize := int64(len(newFileData))

		// create hash for file content
		hashText, err := getHashOfFile(newFileData)
		if err != nil {
			loggermdl.LogError("failed to insert file in pack: ", err)
			return err
		}
		// write content to file
		byteWritten, err := f.WriteAt((newFileData), indexStartOffset)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		if byteWritten < len(newFileData) {
			loggermdl.LogError("short write: ", err)
			return err
		}

		// append new record to indext table
		indexTableRecords = createNewIndexRecord(indexTableRecords, filePath, indexStartOffset, newFileSize, hashText)

		// increment startOffset with file size
		indexStartOffset += newFileSize
	}
	err = f.Sync()
	if err != nil {
		loggermdl.LogError("failed to insert file in pack: ", err)
		return errormdl.Wrap("failed to insert file in pack" + err.Error())
	}
	err = f.Close()
	if err != nil {
		loggermdl.LogError(err)
		return errormdl.Wrap("failed to insert file in pack" + err.Error())
	}
	err = wrapUpSession(indexStartOffset, []byte(indexTableRecords), packPath)
	if err != nil {
		loggermdl.LogError("failed to insert file in pack: ", err)
		return errormdl.Wrap("failed to insert file in pack" + err.Error())
	}

	mutex.Unlock()

	return nil
}

// GetFileDataListFromPack - GetFileDataListFromPack
func GetFileDataListFromPack(packPath string, fileNameList []string) (map[string][]byte, error) {
	mapToReturn := make(map[string][]byte)
	f, err := filemdl.Open(packPath)
	if err != nil {
		return mapToReturn, err
	}
	defer f.Close()
	indexData, _, err := fetchIndexTableMeta(f)
	if err != nil {
		return mapToReturn, err
	}

	for ind := 0; ind < len(fileNameList); ind++ {
		sizeOfFile := gjson.Get(indexData, `#[fileName=`+fileNameList[ind]+`].allocation.byteSize`)
		startRead := gjson.Get(indexData, `#[fileName=`+fileNameList[ind]+`].allocation.startOffset`)
		_, err = f.Seek(startRead.Int(), 0)
		if err != nil {
			continue
		}
		bytesOfFile := make([]byte, sizeOfFile.Int())
		_, err = f.Read(bytesOfFile)
		if err != nil {
			continue
		}
		mapToReturn[fileNameList[ind]] = bytesOfFile
	}

	return mapToReturn, nil
}

// GetFileDataFromPack - GetFileDataFromPack
func GetFileDataFromPack(packPath string, fileName string) ([]byte, error) {
	f, err := filemdl.Open(packPath)
	if err != nil {
		return []byte{}, err
	}
	defer f.Close()
	indexData, _, err := fetchIndexTableMeta(f)
	if err != nil {
		return []byte{}, err
	}
	// read data from pack
	sizeOfFile := gjson.Get(indexData, `#[fileName=`+fileName+`].allocation.byteSize`)
	startRead := gjson.Get(indexData, `#[fileName=`+fileName+`].allocation.startOffset`)
	_, err = f.Seek(startRead.Int(), 0)
	if err != nil {
		return []byte{}, err
	}
	bytesOfFile := make([]byte, sizeOfFile.Int())
	_, err = f.Read(bytesOfFile)
	if err != nil {
		return []byte{}, err
	}
	return bytesOfFile, nil
}

// DeleteFileFromPack - deletes file from pack
func DeleteFileFromPack(packPath string, fileName string) error {
	mutex := sync.Mutex{}
	mutex.Lock()
	f, err := filemdl.Open(packPath)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	defer f.Close()
	indexTableRecords, indexStartOffset, err := fetchIndexTableMeta(f)
	if err != nil {
		loggermdl.LogError("failed to delete file from pack: ", err)
		return err
	}
	indexRecords := gjson.Parse(indexTableRecords)
	if !indexRecords.IsArray() {
		loggermdl.LogError("no file found to delete")
		return errormdl.Wrap("no file found to delete")
	}

	indexToDelete := -1
	for index, record := range indexRecords.Array() {
		if fileName == record.Get("fileName").String() {
			indexToDelete = index
			break
		}
	}
	if indexToDelete == -1 {
		loggermdl.LogError("no file found to delete")
		return errormdl.Wrap("no file found to delete")
	}

	updatedIndexTableRecords, err := sjson.Delete(indexTableRecords, strconv.Itoa(indexToDelete))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	err = wrapUpSession(indexStartOffset, []byte(updatedIndexTableRecords), packPath)
	if err != nil {
		loggermdl.LogError("failed to delete file from pack: ", err)
		return errormdl.Wrap("failed to delete file from pack" + err.Error())
	}
	mutex.Unlock()

	return nil

}

// SearchPackFilePath - SearchPackFilePath
func SearchPackFilePath(dirPath, searchTillPath string) (string, error) {
	path := ""
	dirPath = filepath.Join(dirPath)
	pathChunkList := strings.Split(strings.Trim(dirPath, string(filepath.Separator)), string(filepath.Separator))
	lenPathChunkList := len(pathChunkList)
	trimmedBucket := ""
	for i := 1; true; i++ {
		fileInfoList, err := filemdl.ListDirectory(dirPath)
		if err != nil {
			loggermdl.LogError("failed to find pack path: ", err)
			return "", errormdl.Wrap("failed to find pack path: " + err.Error())
		}

		for _, fileInfo := range fileInfoList {
			if trimmedBucket+".pack" == fileInfo.Name() {
				path = filepath.Join(dirPath, fileInfo.Name())
				break
			}
		}
		if len(dirPath) == 0 || dirPath == searchTillPath {
			break
		}

		trimmedBucket = pathChunkList[lenPathChunkList-i]
		dirPath = strings.TrimSuffix(dirPath, string(filepath.Separator)+trimmedBucket)

	}
	if len(path) == 0 {
		return "", errormdl.Wrap(PackNotFound)
	}
	return path, nil
}

func fetchIndexTableMeta(file *os.File) (string, int64, error) {
	_, err := file.Seek(0, 0)
	if err != nil {
		return "", -1, err
	}
	indexData := make([]byte, 15)
	_, err = file.Read(indexData)
	if err != nil {
		return "", -1, err
	}
	indexDataStartHeader := string(indexData)
	_, err = file.Seek(15, 0)
	if err != nil {
		return "", -1, err
	}
	indexSize := make([]byte, 15)
	_, err = file.Read(indexSize)
	if err != nil {
		return "", -1, err
	}
	startOffset, err := strconv.ParseInt(indexDataStartHeader, 10, 64)
	indexDataSizeHeader := string(indexSize)
	start, _ := strconv.Atoi(indexDataStartHeader)
	size, _ := strconv.Atoi(indexDataSizeHeader)
	_, err = file.Seek(int64(start), 0)
	if err != nil {
		return "", -1, err
	}
	metaByte := make([]byte, size)
	_, err = file.Read(metaByte)
	if err != nil {
		return "", -1, err
	}

	decryptedData, err := securitymdl.AESDecrypt(metaByte, []byte(encryptionKey))
	if err != nil {
		loggermdl.LogError(err)
		return "", -1, err
	}
	return string(decryptedData), startOffset, nil
}

func fetchIndexTable(packFilePath string) (string, int64, error) {
	f, err := filemdl.Open(packFilePath)
	if err != nil {
		return "", 0, errormdl.Wrap("no pack found: " + packFilePath)
	}
	defer f.Close()
	return fetchIndexTableMeta(f)
}

func getFileSizeFromPack(data string, name string) int64 {
	sizeOfFile := gjson.Get(data, `#[fileName=`+name+`].allocation.byteSize`)
	return (sizeOfFile.Int())
}

func getFileStartOffsetFromPack(data string, name string) int64 {
	startoffset := gjson.Get(data, `#[fileName=`+name+`].allocation.startOffset`)
	return (startoffset.Int())
}

func getFileHashTextFromPack(data string, name string) string {
	hashText := gjson.Get(data, `#[fileName=`+name+`].allocation.hashText`)
	return hashText.String()
}

func appendPaddingToNumber(value int64) string {
	return fmt.Sprintf("%015d", value)
}

func wrapUpSession(endOffset int64, indexingData []byte, packCreationPath string) error {
	f, err := filemdl.OpenFile(packCreationPath,
		os.O_CREATE|os.O_WRONLY, 0644)
	if err != nil {
		return err
	}
	defer f.Close()
	incryptedIndexData, err := securitymdl.AESEncrypt(indexingData, []byte(encryptionKey))
	if err != nil {
		return err
	}
	f.WriteAt(incryptedIndexData, endOffset)
	f.WriteAt([]byte(appendPaddingToNumber(endOffset)), 0)
	f.WriteAt([]byte(appendPaddingToNumber(int64(len(incryptedIndexData)))), 15)
	return nil
}

func updateJSON(jsonData string, name string, startOffset int64, size int64, hashText string) string {
	res := gjson.Parse(jsonData)
	var updatedRes []interface{}

	res.ForEach(func(key, value gjson.Result) bool {
		// fmt.Println("val:", value.Get("Filename").String())
		if value.Get("fileName").String() == name {
			// Update byte size
			updateValue, err := sjson.Set(value.String(), "allocation.byteSize", size)
			if err != nil {
				return false
			}
			// Update startOffset
			updateValue, err = sjson.Set(updateValue, "allocation.startOffset", startOffset)
			if err != nil {
				return false
			}
			// Update hashText
			updateValue, err = sjson.Set(updateValue, "allocation.hashText", hashText)
			if err != nil {
				return false
			}
			updatedRes = append(updatedRes, gjson.Parse(updateValue).Value())
		} else {
			updatedRes = append(updatedRes, value.Value())
		}
		return true // keep iterating
	})
	temp := `{}`
	updateValue, _ := sjson.Set(temp, "writeData", updatedRes)
	fileData := gjson.Parse(updateValue)

	return fileData.Get("writeData").String()
}

func getHashOfFile(data []byte) (string, error) {
	// create uint64 hash for file content
	hashOfSourceData, err := hashmdl.GetHashChecksumOfByteArray(data)
	if err != nil {
		return "", err
	}
	// convert uint64 hash to string
	hashText := strconv.FormatUint(hashOfSourceData, 10)
	return hashText, nil
}

func createNewIndexRecord(indexTableRecords string, Filename string, startoffset int64, bytesize int64, hashText string) string {
	jsonzobj := `{}`
	jsonzobj, _ = sjson.Set(jsonzobj, "fileName", Filename)
	jsonzobj, _ = sjson.Set(jsonzobj, "allocation.startOffset", startoffset)
	jsonzobj, _ = sjson.Set(jsonzobj, "allocation.byteSize", bytesize)
	jsonzobj, _ = sjson.Set(jsonzobj, "allocation.hashText", hashText)
	parsedJsonzObj := gjson.Parse(jsonzobj)
	indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())

	return indexTableRecords
}
